function [t, R, x, w, v, fw, fv] = rb_step(t, R, x, w, v, h, S)
% Dynamics update of a rigid body with basic stabilization control laws
%
% @param t - time (scalar)
% @param R - orientation matrix (3x3 matrix)
% @param x - position (3x1 vector)
% @param w - angular velocity (3x1 vector)
% @param v - linear velocity (3x1 vector)
% @param h - time step (scalar)
% @param S - system parameters structure given by:
%     S.J - principal moments of inertia vector (3x1 vector)
%     S.m - mass (scalar)
%     S.Kwp - proportional angular gains (3x1 vector)
%     S.Kwd - derivative angular gains (3x1 vector)
%     S.Kvp - proportional linear gains (3x1 vector)
%     S.Kvd - derivative linear gains (3x1 vector)
%     S.R0  - desired orientation (3x3 matrix)
%     S.x0  - desired position (3x1 vector)
%     S.fwlb - torque lower bound (3x1 vector)
%     S.fwub - torque upper bound (3x1 vector)
%     S.fvlb - force lower bound (3x1 vector)
%     S.fvub - force upper bound (3x1 vector)
%     S.fctl - control torques and forces, given by a handle
%              to a function of the form:
%              function [fw, fv] = fctl(t, R, x, w, v, S) 
%     S.fext - any external torques and forces, given by a handle
%              to a function of the form:
%              function [fwe, fve] = fext(t, R, x, w, v, S) 
%     S.fint - any internal torques and forces (technically same as
%              fctl, but given separately for convenience), given by a handle
%              to a function of the form:
%              function [fwi, fvi] = fint(t, R, x, w, v, S) 
% Note: all parameters are optional
%
% @return t - updated time (scalar)
% @return R - updated orientation matrix (3x3 matrix)
% @return w - updated angular velocity (3x1 vector)
% @return fw - computed torque (3x1 vector)
%
% Author: Marin Kobilarov, marin(at)cds.caltech.edu

% set defaults
if ~isfield(S, 'J')
  S.J = [1; 1; 1];
end

if ~isfield(S, 'Kwp')
  S.Kwp = [.1; .1; .1];
end

if ~isfield(S, 'Kwd')
  S.Kwd = [1; 1; 1];
end

if ~isfield(S, 'Kvp')
  S.Kvp = [.1; .1; .1];
end

if ~isfield(S, 'Kvd')
  S.Kvd = [1;1;1];
end

if ~isfield(S, 'R0')
  S.R0 = eye(3);
end

if ~isfield(S, 'x0')
  S.x0 = zeros(3,1);
end


if ~isfield(S, 'fwlb')
  S.fwlb = -[inf; inf; inf];
end

if ~isfield(S, 'fwub')
  S.fwub = [inf; inf; inf];
end

if ~isfield(S, 'fvlb')
  S.fvlb = -[inf; inf; inf];
end

if ~isfield(S, 'fvub')
  S.fvub = [inf; inf; inf];
end

if ~isfield(S, 'fctl')
  S.fctl = @fctl;
end

if ~isfield(S, 'fext')
  S.fext = @fext;
end

if ~isfield(S, 'fint')
  S.fint = @fint;
end


% control forces
[fw, fv] = feval(S.fctl, t, R, x, w, v, S);

% bound it
fw = min(fw, S.fwub);
fw = max(fw, S.fwlb);

fv = min(fv, S.fvub);
fv = max(fv, S.fvlb);


% external forces
[fwe, fve] = feval(S.fext, t, R, x, w, v, S);

% internal forces
[fwi, fvi] = feval(S.fint, t, R, x, w, v, S);
fw = fw + fwi;
fv = fv + fvi;

% update
t = t + h;

euler = 1;
if euler,
  % standard Euler -- degrades rapidly over time
  w = w + (h./S.J).*(cross(S.J.*w, w) + fw + R'*fwe);
else
  % symplectic trapezoidal variational integrator
  a = (eye(3) - 1/2*so3_ad(h*w))*(S.J.*w) + h*(fw + R'*fwe);

  opts = optimset('Display','off');
%  opts = optimset('Display','off','TolCon', 1e-10, 'TolFun', 1e-10, 'TolX', 1e-10);
  [w, fval, ef, output] = fsolve(@rb_dyn, w, opts, h, a, S);
  if (ef < 1)
    disp('attitude dynamics could not converge!')
  end
end

R = R*so3_exp(h*w);

v = v + (h/S.m)*(R*fv + fve);
x = x + h*v;


function f = rb_dyn(w, h, a, S)
f = (eye(3) + 1/2*so3_ad(h*w))*(S.J.*w) - a;


function [fw, fv] = fctl(t, R, x, w, v, S)
% default external forces

fw = zeros(3,1);
fv = zeros(3,1);


function [fwe, fve] = fext(t, R, x, w, v, S)
% default external forces

fwe = zeros(3,1);
fve = zeros(3,1);


function [fwe, fve] = fint(t, R, x, w, v, S)
% default external forces

fwe = zeros(3,1);
fve = zeros(3,1);