function f = rb_test
% Simulation of a basic fully-actuated rigid-body control law
% stabilizing the system to a desired equilibirum. The key
% features are an dynamics integrator which uses directly rotation 
% matrices, rather than Euler angles, and similary a control law 
% which uses the logarithm map on SO(3) to compute rotation errors.
% 
% EN530.678: Nonlinear control and planning in robotics
% M. Kobilarov,  marin(at)jhu.edu

clear all;

% satellite dimension
d = [.1; .1; .34];

% mass
S.m = 3.4;

% moments of inertia
S.J = S.m/12*[d(2)^2+d(3)^2; d(1)^2+d(3)^2; d(1)^2+d(2)^2];

% initial state
t = 0;
x = [-5; -3; 5];
R = so3_exp([pi/4; pi/8; pi/2]);
%R = so3_exp([0; 0; 0]);
w = [0;0;0];
v = [-.05;-.05;0];


% gains
Ts  = 1;   % time-scale
S.Kwp = 1/(Ts*Ts)*[1;1;1];
S.Kwd = 1/(Ts)*[5;5;5];

S.Kvp = 1/(Ts*Ts)*[1;1;1];
S.Kvd = 1/(Ts)*[5;5;5];

% controller
S.fctl = @fctl_stab;


% number of time steps
N = 500;

% total time (seconds)
T = 30*Ts;

h=T/N;
S.h = h;

% desired state
S.R0 = eye(3);
S.w0 = zeros(3,1);
S.x0 = zeros(3,1);
S.v0 = zeros(3,1);

% arrays to store the output
es = zeros(3, N);
xs = zeros(3, N);
ws = zeros(3, N);
vs = zeros(3, N);
fws = zeros(3, N);
fvs = zeros(3, N);

% graphics handle
G = [];
hold on
draw_rb([], S.x0, S.R0, 'g', d);

axis equal; axis tight; axis manual;

view(3)

for i=1:N,
  i
  G = draw_rb(G, x, R, 'b', d);
  axis([-2 2 -2 2 -2 2])
  drawnow
  
  [t, R, x, w, v, fw, fv] = rb_step(t, R, x, w, v, h, S);
  
  % log the output
  eRs(:,i) = so3_log(S.R0'*R);
  exs(:,i) = x - S.x0;
  ws(:,i) = w;
  vs(:,i) = v;
  fws(:,i) = fw;
  fvs(:,i) = fv;
  
  pause(h/2);
end

% display output

figure
set(gca, 'FontSize',20);


ts = 0:h:(N-1)*h;

subplot(3,2,1)
plot(ts,eRs','LineWidth',2)
legend('R_x','R_y','R_z')
xlabel('Time (s)','FontSize',20)
ylabel('Angles (rad)','FontSize',20)
%strtitle = strcat('Attitude error - Actuator bound:',num2str(S.fwub(1)*1000),'mN');
%title(strtitle,'FontSize',20)
set(gca, 'FontSize',20);

subplot(3,2,2)
plot(exs','LineWidth',2)
legend('x','y','z')
xlabel('time')
ylabel('m')
title('Position error')
set(gca, 'FontSize',15);

subplot(3,2,3)
plot(ws','LineWidth',2)
legend('w_x','w_y','w_z')
xlabel('time')
ylabel('rad/s')
title('Attitude Rate')
set(gca, 'FontSize',15);

subplot(3,2,4)
plot(vs','LineWidth',2)
legend('v_x','v_y','v_z')
xlabel('time')
ylabel('m/s')
title('Velocity')
set(gca, 'FontSize',15);

subplot(3,2,5)
plot(fws','LineWidth',2)
legend('fw_x','fw_y','fw_z')
xlabel('time')
ylabel('N/m')
title('Torque')
set(gca, 'FontSize',15);

subplot(3,2,6)
plot(fvs','LineWidth',2)
legend('fv_x','fv_y','fv_z')
xlabel('time')
ylabel('N')
title('Forces')
set(gca, 'FontSize',15);

drawnow

set(gcf, 'OuterPosition', [100 100 600 600])

%set(gcf, 'PaperPosition', [100 200 800 400]);



function [fw, fv] = fctl_stab(t, R, x, w, v, S) 
% default control law assuming full actuation

% take the error on SO(3)
we = so3_log(R'*S.R0);

% take the difference in position
ve = R'*(S.x0 - x);

% desired force (prop + deriv + remove cross-coupling)
fw = S.J.*(S.Kwp.*we - S.Kwd.*w) - cross(S.J.*w, w);

fv = S.m*(S.Kvp.*ve - S.Kvd.*(R'*v));

if ~isfield(S, 'fctl')
  S.fctl = @fctl;
end